package ru.dbotthepony.mc.otm.data

import com.google.gson.JsonObject
import com.google.gson.JsonParseException
import com.google.gson.JsonSyntaxException
import com.mojang.datafixers.util.Pair
import com.mojang.serialization.Codec
import com.mojang.serialization.DataResult
import com.mojang.serialization.DynamicOps
import com.mojang.serialization.JsonOps
import io.netty.buffer.ByteBuf
import io.netty.buffer.Unpooled
import net.minecraft.advancements.AdvancementHolder
import net.minecraft.data.recipes.FinishedRecipe
import net.minecraft.network.FriendlyByteBuf
import net.minecraft.resources.ResourceLocation
import net.minecraft.world.item.ItemStack
import net.minecraft.world.item.crafting.Ingredient
import net.minecraft.world.item.crafting.Recipe
import net.minecraft.world.item.crafting.RecipeSerializer
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.mc.otm.core.set
import ru.dbotthepony.mc.otm.core.util.readBinaryJsonWithCodecIndirect
import ru.dbotthepony.mc.otm.core.util.writeBinaryJsonWithCodec
import kotlin.collections.ArrayDeque
import kotlin.concurrent.getOrSet
import kotlin.reflect.KProperty

/**
 * 1.20.2: Mojang FINALLY moved json IO of recipes to codecs
 *
 * ...but they forgot to do the same for Networking them, Ingredient does not have codec for network.
 *
 * Mojang.... Mojang never changes.
 */
class Codec2RecipeSerializer<S : Recipe<*>>(
	val empty: S?,
	codec: (Codec2RecipeSerializer<S>.Context) -> Codec<S>,
) : Codec<S>, RecipeSerializer<S> {
	constructor(supplier: (Codec2RecipeSerializer<S>.Context) -> Codec<S>) : this(null, supplier)

	private class CurrentContext {
		var isNetwork = 0
	}

	inner class Context {
		val ingredients: Codec<Ingredient> get() = ActualIngredientCodec

		fun <P : Recipe<*>> wrap(serializer: RecipeSerializer<P>): Codec<P> {
			return object : Codec<P> {
				override fun <T : Any?> encode(input: P, ops: DynamicOps<T>, prefix: T): DataResult<T> {
					if (context.isNetwork > 0) {
						val parent = Unpooled.buffer()
						val buff = FriendlyByteBuf(parent)
						serializer.toNetwork(buff, input)
						return DataResult.success(ops.createByteList(parent.nioBuffer()))
					} else {
						return serializer.codec().encode(input, ops, prefix)
					}
				}

				override fun <T : Any?> decode(ops: DynamicOps<T>, input: T): DataResult<Pair<P, T>> {
					if (context.isNetwork > 0) {
						return ops.getByteBuffer(input).flatMap {
							val parent = Unpooled.buffer()
							val buff = FriendlyByteBuf(parent)
							parent.writeBytes(it)
							parent.setIndex(0, 0)
							val read = serializer.fromNetwork(buff)

							if (read == null)
								DataResult.error { "Unable to read parent recipe from network" }
							else
								DataResult.success(Pair(read, ops.empty()))
						}
					} else {
						return serializer.codec().decode(ops, input)
					}
				}
			}
		}
	}

	private val codec = codec.invoke(Context())

	override fun <T : Any> encode(input: S, ops: DynamicOps<T>, prefix: T): DataResult<T> {
		return codec.encode(input, ops, prefix)
	}

	override fun <T : Any> decode(ops: DynamicOps<T>, input: T): DataResult<Pair<S, T>> {
		return codec.decode(ops, input)
	}

	fun <O : Recipe<*>> xmap(to: (S) -> O, from: (O) -> S): Codec2RecipeSerializer<O> {
		return Codec2RecipeSerializer(empty?.let(to)) { _ ->
			codec.xmap(to, from)
		}
	}

	override fun codec(): Codec<S> {
		return this
	}

	override fun fromNetwork(data: FriendlyByteBuf): S? {
		try {
			context.isNetwork++
			return data.readBinaryJsonWithCodecIndirect(this)
				.resultOrPartial { LOGGER.error("Failed to read recipe from network: $it") }.orElse(null)
		} finally {
			context.isNetwork--
		}
	}

	override fun toNetwork(data: FriendlyByteBuf, recipe: S) {
		try {
			context.isNetwork++
			data.writeBinaryJsonWithCodec(this, recipe)
		} finally {
			context.isNetwork--
		}
	}

	fun toFinished(recipe: S, id: ResourceLocation): FinishedRecipe {
		return object : FinishedRecipe {
			override fun serializeRecipeData(p_125967_: JsonObject) {
				encode(recipe, JsonOps.INSTANCE, p_125967_).get().map(
					{
						it as JsonObject

						for ((k, v) in it.entrySet()) {
							p_125967_[k] = v
						}
					},
					{
						throw JsonParseException("Failed to serialize recipe: ${it.message()}")
					}
				)
			}

			override fun id(): ResourceLocation {
				return id
			}

			override fun type(): RecipeSerializer<*> {
				return this@Codec2RecipeSerializer
			}

			override fun advancement(): AdvancementHolder? {
				return null
			}
		}
	}

	private object ActualIngredientCodec : Codec<Ingredient> {
		override fun <T : Any> encode(input: Ingredient, ops: DynamicOps<T>, prefix: T): DataResult<T> {
			return if (context.isNetwork > 0) {
				networkIngredientCodec.encode(input, ops, prefix)
			} else {
				Ingredient.CODEC.encode(input, ops, prefix)
			}
		}

		override fun <T : Any> decode(ops: DynamicOps<T>, input: T): DataResult<Pair<Ingredient, T>> {
			return if (context.isNetwork > 0) {
				networkIngredientCodec.decode(ops, input)
			} else {
				Ingredient.CODEC.decode(ops, input)
			}
		}
	}

	companion object {
		private val LOGGER = LogManager.getLogger()
		private val networkIngredientCodec = Codec.list(ItemStack.CODEC).xmap({ Ingredient.of(it.stream()) }, { it.items.toMutableList() })

		/**
		 * [ThreadLocal] because optimization mods can (and probably should) parallelize recipe deserialization,
		 * since RecipeSerializers are expected to be stateless. [Codec2RecipeSerializer], however, is stateful (threading PoV).
		 * To make it stateless, [ThreadLocal] is used.
		 */
		private val context by object : ThreadLocal<CurrentContext>() {
			override fun initialValue(): CurrentContext {
				return CurrentContext()
			}
		}
	}
}

private operator fun <T> ThreadLocal<T>.getValue(companion: Codec2RecipeSerializer.Companion, property: KProperty<*>): T {
	return get()
}
