package ru.dbotthepony.mc.otm.registry.objects

import com.google.common.collect.Streams
import net.minecraft.core.BlockPos
import net.minecraft.world.item.BlockItem
import net.minecraft.world.item.DyeColor
import net.minecraft.world.item.Item
import net.minecraft.world.level.BlockGetter
import net.minecraft.world.level.block.Block
import net.minecraft.world.level.block.state.BlockBehaviour
import net.minecraft.world.level.block.state.BlockState
import net.minecraft.world.phys.shapes.CollisionContext
import net.minecraft.world.phys.shapes.VoxelShape
import net.minecraftforge.registries.DeferredRegister
import net.minecraftforge.registries.RegistryObject
import ru.dbotthepony.mc.otm.block.RotatableMatteryBlock
import ru.dbotthepony.mc.otm.block.getShapeForEachState
import ru.dbotthepony.mc.otm.core.collect.SupplierMap
import ru.dbotthepony.mc.otm.core.get
import ru.dbotthepony.mc.otm.core.math.BlockRotationFreedom
import ru.dbotthepony.mc.otm.core.util.WriteOnce
import ru.dbotthepony.mc.otm.registry.MRegistry
import ru.dbotthepony.mc.otm.shapes.BlockShape
import ru.dbotthepony.mc.otm.shapes.BlockShapes
import java.util.stream.Stream

/**
 * Base + Colored
 */
class DecorativeBlock(
	baseName: String,
	private val provider: (DyeColor?) -> Block,
) : ColoredDecorativeBlock(baseName, provider) {
	private var _block: RegistryObject<Block> by WriteOnce()
	private var _item: RegistryObject<Item> by WriteOnce()

	val block: Block get() = _block.get()
	val item: Item get() = _item.get()

	val allBlocks: Map<DyeColor?, Block> by lazy {
		check(registeredBlocks) { "Didn't register items yet" }
		SupplierMap(Streams.concat(MRegistry.DYE_ORDER.stream().map { it to blockMap[it]!! }, Stream.of(null to _block)))
	}

	val allItems: Map<DyeColor?, Item> by lazy {
		check(registeredItems) { "Didn't register items yet" }
		SupplierMap(Streams.concat(MRegistry.DYE_ORDER.stream().map { it to itemMap[it]!! }, Stream.of(null to _item)))
	}

	override fun registerBlocks(registry: DeferredRegister<Block>) {
		_block = registry.register(baseName) { provider.invoke(null) }
		super.registerBlocks(registry)
	}

	override fun registerItems(registry: DeferredRegister<Item>) {
		_item = registry.register(baseName) { BlockItem(_block.get(), Item.Properties().stacksTo(64)) }
		super.registerItems(registry)
	}

	companion object {
		fun simple(baseName: String, provider: (DyeColor?) -> BlockBehaviour.Properties): DecorativeBlock {
			return DecorativeBlock(baseName) {
				Block(provider.invoke(it))
			}
		}

		fun rotatable(baseName: String, provider: (DyeColor?) -> BlockBehaviour.Properties): DecorativeBlock {
			return DecorativeBlock(baseName) {
				RotatableMatteryBlock(provider.invoke(it))
			}
		}

		fun rotatable(baseName: String, shape: BlockShape, rotationFreedom: BlockRotationFreedom, provider: (DyeColor?) -> BlockBehaviour.Properties): DecorativeBlock {
			return DecorativeBlock(baseName) {
				object : RotatableMatteryBlock(provider.invoke(it)) {
					override fun rotationFreedom(): BlockRotationFreedom {
						return rotationFreedom
					}

					private val shapes = getShapeForEachState(rotationProperty) { shape.rotateFromNorth(it[rotationProperty]).computeShape() }

					@Suppress("override_deprecation")
					override fun getShape(
						state: BlockState,
						blockGetter: BlockGetter,
						pos: BlockPos,
						context: CollisionContext
					): VoxelShape {
						return shapes[state]!!
					}
				}
			}
		}
	}
}
